import multiprocessing
from typing import Dict, Tuple

from pystan import StanModel

from bqme._settings import STAN_TEMPLATE_PATH
from bqme.distributions import Distribution, Normal, Gamma, Lognormal, Weibull
from bqme.fit_object import FitObjectSampling, FitObjectOptimizing


multiprocessing.set_start_method("fork") #mac has diffrerent default


class QM:
    """
    Base class for quantile matching (QM) models

    Parameters
    __________
    parameters_dict : Dict
        Keys are internal names for the priors of the model
        e.g. 'mu', 'sigma' for a GaussianQM. Values are the user 
        defined Distributions. Note key must not be identical to value.name.
    """
    def __init__(self, parameters_dict: Dict[str, Distribution]):
        self.parameters_dict = self._check_dict(parameters_dict)
        self.model = None

    def __str__(self):
        return self.__class__.__name__ + '(' +  \
            ', '.join([p.__str__() for p in self.parameters_dict.values()]) + \
            ')'

    def __repr__(self):
        return self.__str__()

    def _check_dict(self, parameters_dict:Dict[str, Distribution]):
        for key, value in parameters_dict.items():
            if not isinstance(value, Distribution):
                raise ValueError(f'Input parameter "{key}" of "{self.__class__.__name__}" needs to be a Distribution (see bqme.distributions), but is of type {type(value)}.')
        return parameters_dict

    def _template_replacements(self) -> Dict['str', 'str']:
        """
        returns a dict that contains keys as template variables
        and values are the strings for the variables.
        Necessary keys: parametersnames, parameters, priors, cdf, lpdf, rng
        """
        distribution_name = self.__class__.__name__.replace("QM", "").lower()
        build = lambda s: '\n    '.join([
                p.code()[s] for p in self.parameters_dict.values()
            ])
        replacements = {
                'parametersnames'   : ', '.join([
                        p.name for p in self.parameters_dict.values()
                    ]),
                'parameters'        : build('parameter'),
                'priors'            : build('prior'),
                'cdf'               : f'{distribution_name}_cdf',
                'lpdf'              : f'{distribution_name}_lpdf',
                'rng'               : f'{distribution_name}_rng',
            }
        return replacements

    def _stan_code(self) -> str:
        with open(STAN_TEMPLATE_PATH) as f:
            code = f.read()
        for k, v in self._template_replacements().items():
            code = code.replace(f'${k}$', v)
        return code

    def _check_domain(self, X):
        minn, maxx = self.domain()
        f = lambda x: not(minn < x < maxx)
        if len(list(filter(f, X))) > 0:
            raise ValueError(f'some elements of X are not in the domain of the model, which is ({minn}, {maxx}).')

    def domain(self): pass

    @property
    def code(self) -> str:
        """ returns the final stan code """
        return self._stan_code()

    def compile(self):
        self.model = StanModel(model_code=self.code)

    def sampling(self, N:int, q:Tuple[float,...], X:Tuple[float,...]) -> 'StanFit4Model':
        self._check_domain(X)
        if self.model is None: self.compile()
        data_dict = {'N':N, 'M':len(q), 'q':q, 'X':X}
        samples = self.model.sampling(data=data_dict)
        return FitObjectSampling(self, samples)

    def optimizing(self, N:int, q:Tuple[float,...], X:Tuple[float,...]) -> 'StanFit4Model':
        self._check_domain(X)
        if self.model is None: self.compile()
        data_dict = {'N':N, 'M':len(q), 'q':q, 'X':X}
        opt = self.model.optimizing(data=data_dict)
        return FitObjectOptimizing(self, opt)


class NormalQM(QM):
    """
    Quantile matching using Normal distribution

    Parameters
    ----------
    mu : Distribution
        location of the Normal
    sigma : Distribution
        scale of the Normal

    Examples
    --------
    >>> from bqme.distributions import Normal, Gamma
    >>> model = NormalQM(Normal(0., 1., 'mu'), Gamma(1., 1., 'sigma'))
    >>> model
    NormalQM(Normal(mu=0.0, sigma=1.0, name="mu"), Gamma(alpha=1.0, beta=1.0, name="sigma"))
    >>> code = model.code
    """
    def __init__(self, mu:Distribution, sigma:Distribution):
        self.mu = mu
        self.sigma = sigma
        self._distribution = Normal #to access corresponding distribution in fit
        parameters_dict = {'mu': self.mu, 'sigma': self.sigma}
        super().__init__(parameters_dict)

    def domain(self) -> Tuple[float, float]:
        return (float('-inf'), float('inf'))


class GammaQM(QM):
    """
    Quantile matching using Gamma distribution

    Parameters
    ----------
    alpha : Distribution
        Also called the shape of the Gamma
    beta : Distribution
        Also called the rate of the Gamma

    Examples
    --------
    >>> from bqme.distributions import Gamma
    >>> model = GammaQM(Gamma(1., 1. ,'alpha'), Gamma(1., 1., 'beta'))
    >>> model
    GammaQM(Gamma(alpha=1.0, beta=1.0, name="alpha"), Gamma(alpha=1.0, beta=1.0, name="beta"))
    >>> code = model.code
    """
    def __init__(self, alpha:Distribution, beta:Distribution):
        self.alpha = alpha
        self.beta = beta
        self._distribution = Gamma #to access corresponding distribution in fit
        parameters_dict = {'alpha': self.alpha, 'beta': self.beta}
        super().__init__(parameters_dict)

    def domain(self) -> Tuple[float, float]:
        return (0, float('inf'))


class LognormalQM(QM):
    """
    Quantile matching using Lognormal distribution

    Parameters
    ----------
    mu : Distribution
        location of the corresponding Normal distribution
    sigma : Distribution
        scale of the corresponding Normal distribution

    Examples
    --------
    >>> from bqme.distributions import Lognormal, Normal
    >>> model = LognormalQM(Normal(1., 1. ,'mu'), Lognormal(1., 1., 'sigma'))
    >>> model
    LognormalQM(Normal(mu=1.0, sigma=1.0, name="mu"), Lognormal(mu=1.0, sigma=1.0, name="sigma"))
    >>> code = model.code
    """
    def __init__(self, mu:Distribution, sigma:Distribution):
        self.mu = mu
        self.sigma = sigma
        self._distribution = Lognormal #to access corresponding distribution in fit
        parameters_dict = {'mu': self.mu, 'sigma': self.sigma}
        super().__init__(parameters_dict)

    def domain(self) -> Tuple[float, float]:
        return (0, float('inf'))


class WeibullQM(QM):
    """
    Quantile matching using Weibull distribution

    Parameters
    ----------
    alpha : Distribution
        Also called the shape of the Weibull
    sigma : Distribution
        Also called the scale of the Weibull

    Examples
    --------
    >>> from bqme.distributions import Weibull
    >>> model = WeibullQM(Weibull(1., 1. ,'alpha'), Weibull(1., 1., 'sigma'))
    >>> model
    WeibullQM(Weibull(alpha=1.0, sigma=1.0, name="alpha"), Weibull(alpha=1.0, sigma=1.0, name="sigma"))
    >>> code = model.code
    """
    def __init__(self, alpha:Distribution, sigma:Distribution):
        self.alpha = alpha
        self.sigma = sigma
        self._distribution = Weibull #to access corresponding distribution in fit
        parameters_dict = {'alpha': self.alpha, 'sigma': self.sigma}
        super().__init__(parameters_dict)

    def domain(self) -> Tuple[float, float]:
        return (0, float('inf'))



import pytest
from bqme.variables import ContinuousVariable, PositiveContinuousVariable

def test_continuousVariable():
    mu = ContinuousVariable(-1., name='mu')
    assert mu.value == -1.
    assert mu.name == 'mu'
    assert mu.upper == float('inf')
    assert mu.lower == float('-inf')

def test_PositiveContinuousVariable():
    sigma = PositiveContinuousVariable(1., name='sigma')
    assert sigma.value == 1.
    assert sigma.name == 'sigma'
    assert sigma.upper == float('inf')
    assert sigma.lower == 0.

def test_PositiveContinuousVariable_expected_fail():
    with pytest.raises(ValueError):
        PositiveContinuousVariable(-1., name='sigma')




import pytest

from bqme.distributions import Normal, Gamma, Lognormal, Weibull
from bqme.models import QM, NormalQM, GammaQM, LognormalQM, WeibullQM


def pytest_addoption(parser):
    parser.addoption("--slow", action="store_true", help="run slow tests")

def pytest_runtest_setup(item):
    if 'slow' in item.keywords and not item.config.getvalue("slow"):
        pytest.skip("need --slow option to run")

##### DEFINE FIXTURES HERE

@pytest.fixture(scope='session')
def normal_compiled_model():
    mu = Normal(0., 1., name='mu')
    sigma = Gamma(1., 1.2, name='sigma')
    model = NormalQM(mu, sigma)
    model.compile()
    return  model

@pytest.fixture(scope='module')
def gamma_compiled_model():
    alpha = Gamma(1., 1., name='alpha')
    beta = Gamma(1., 1., name='beta')
    model = GammaQM(alpha, beta)
    model.compile()
    return  model

@pytest.fixture(scope='module')
def lognormal_compiled_model():
    mu = Normal(1., 1., name='mu')
    sigma = Lognormal(1., 1., name='sigma')
    model = LognormalQM(mu, sigma)
    model.compile()
    return  model

@pytest.fixture(scope='module')
def weibull_compiled_model():
    alpha = Weibull(1., 1., name='alpha')
    sigma = Weibull(1., 1., name='sigma')
    model = WeibullQM(alpha, sigma)
    model.compile()
    return  model




import pytest
import numpy as np
import pystan
from scipy.stats import norm, gamma, lognorm, weibull_min

from bqme.distributions import Normal, Gamma, Lognormal, Weibull

distributions = [
    (Normal(mu=1., sigma=2., name='n'), norm(loc=1., scale=2.)),
    (Gamma(alpha=1., beta=2., name='g'), gamma(a=1., scale=1./2.)),
    (Lognormal(mu=1., sigma=2., name='l'), lognorm(s=2., scale=np.exp(1.))),
    (Weibull(alpha=1., sigma=2., name='w'), weibull_min(c=1., scale=2.)),
]

### test parameter equivalence of scipy and stan
code = lambda dist_name: f"""parameters {{real x;}}
transformed parameters {{real lprob = {dist_name}_lpdf(x| 2., 0.8);}}
model {{x ~ {dist_name}(2.,0.8);}}"""

def test_normal_print():
    mu = Normal(0, 1, name='mu')
    mu2 = Normal(1., 2, name='mu2')
    assert str(mu) == 'Normal(mu=0, sigma=1, name="mu")'
    assert str(mu2) == 'Normal(mu=1.0, sigma=2, name="mu2")'

def test_normal_code():
    code = Normal(0., 1., name='mu').code()
    exp_out = {
            'parameter': 'real mu;', 
            'prior': 'mu ~ normal(0.0, 1.0);'
        }
    assert code == exp_out

@pytest.mark.slow
def test_normal_parameters():
    model_stan = pystan.StanModel(model_code=code('normal'))
    samples = model_stan.sampling()
    s = samples.extract('x')['x']
    s_lp = samples.extract('lprob')['lprob']
    model_scipy = norm(loc=2., scale=0.8)
    s_lp2 = model_scipy.logpdf(s)
    assert np.allclose(s_lp, s_lp2)

def test_gamma_print():
    alpha = Gamma(1, 1, name='alpha')
    beta = Gamma(1, 1.2, name='somethingelse')
    assert str(alpha) == 'Gamma(alpha=1, beta=1, name="alpha")'
    assert str(beta) == 'Gamma(alpha=1, beta=1.2, name="somethingelse")'

def test_gamma_code():
    code = Gamma(1, 1., name='beta').code()
    exp_out = {
            'parameter': 'real<lower=0> beta;', 
            'prior': 'beta ~ gamma(1, 1.0);'
        }
    assert code == exp_out

@pytest.mark.slow
def test_gamma_parameters():
    model_stan = pystan.StanModel(model_code=code('gamma'))
    samples = model_stan.sampling()
    s = samples.extract('x')['x']
    s_lp = samples.extract('lprob')['lprob']
    model_scipy = gamma(a=2., scale=1./0.8)
    s_lp2 = model_scipy.logpdf(s)
    assert np.allclose(s_lp, s_lp2)

def test_lognormal_print():
    mu = Lognormal(1, 1, name='mu')
    sigma = Lognormal(1, 1.2, name='s')
    assert str(mu) == 'Lognormal(mu=1, sigma=1, name="mu")'
    assert str(sigma) == 'Lognormal(mu=1, sigma=1.2, name="s")'
    assert repr(mu) == 'Lognormal(mu=1, sigma=1, name="mu")'
    assert repr(sigma) == 'Lognormal(mu=1, sigma=1.2, name="s")'

def test_lognormal_code():
    code = Lognormal(1, 1., name='sigma').code()
    exp_out = {
            'parameter': 'real<lower=0> sigma;',
            'prior': 'sigma ~ lognormal(1, 1.0);'
        }
    assert code == exp_out

@pytest.mark.slow
def test_lognormal_parameters():
    model_stan = pystan.StanModel(model_code=code('lognormal'))
    samples = model_stan.sampling()
    s = samples.extract('x')['x']
    s_lp = samples.extract('lprob')['lprob']
    model_scipy = lognorm(s=0.8, scale=np.exp(2.))
    s_lp2 = model_scipy.logpdf(s)
    assert np.allclose(s_lp, s_lp2)

def test_weibull_print():
    alpha = Weibull(1., 1., name='alpha')
    sigma = Weibull(1., 1.2, name='s')
    assert str(alpha) == 'Weibull(alpha=1.0, sigma=1.0, name="alpha")'
    assert str(sigma) == 'Weibull(alpha=1.0, sigma=1.2, name="s")'

def test_weibull_code():
    code = Weibull(1., 1., name='sigma').code()
    exp_out = {
            'parameter': 'real<lower=0> sigma;', 
            'prior': 'sigma ~ weibull(1.0, 1.0);'
        }
    assert code == exp_out

@pytest.mark.slow
def test_weibull_parameters():
    model_stan = pystan.StanModel(model_code=code('weibull'))
    samples = model_stan.sampling()
    s = samples.extract('x')['x']
    s_lp = samples.extract('lprob')['lprob']
    model_scipy = weibull_min(c=2., scale=0.8)
    s_lp2 = model_scipy.logpdf(s)
    assert np.allclose(s_lp, s_lp2)

### EXPECTED FAILS

def test_wrong_initialization():
    with pytest.raises(ValueError):
        Normal(0, 0, name='mu')
    with pytest.raises(ValueError):
        Gamma(-1., 2, name='alpha')


### pdf
@pytest.mark.parametrize("bqme_dist, scipy_dist", distributions)
def test_distribution_scipy(bqme_dist, scipy_dist):
    """ test pdf, cdf, logpdf, logcdf, ppf """
    x = np.linspace(1., 3., 10)
    q = np.linspace(0.1, 0.9, 10)
    assert all( bqme_dist.pdf(x) == scipy_dist.pdf(x) )
    assert all( bqme_dist.cdf(x) == scipy_dist.cdf(x) )
    assert all( bqme_dist.logpdf(x) == scipy_dist.logpdf(x) )
    assert all( bqme_dist.logcdf(x) == scipy_dist.logcdf(x) )
    assert all( bqme_dist.ppf(q) == scipy_dist.ppf(q) )




class Variable:
    """
    Base class for different kind of variables
    """
    def __init__(self, value:float, name:str, lower:float, upper:float):
        self.lower = lower
        self.upper = upper
        self.name = name
        self.value = self.approve(value)

    def approve(self, value:float) -> float:
        if self.lower < value < self.upper:
            return value
        else:
            raise ValueError(f'Input parameter "{self.name}" needs to be in range ({self.lower}, {self.upper}), currently set to {value}.')


class ContinuousVariable(Variable):
    """
    container for continuous variables (-inf, inf)
    """
    def __init__(self, value:float, name:str):
        lower = float('-inf')
        upper = float('inf')
        super().__init__(value, name, lower, upper)


class PositiveContinuousVariable(Variable):
    """
    container for positive continous variables (0, inf)
    """
    def __init__(self, value:float, name:str):
        lower = 0.
        upper = float('inf')
        super().__init__(value, name, lower, upper)





import pytest
import numpy as np

from bqme.fit_object import FitObjectSampling, FitObjectOptimizing

@pytest.mark.slow
def test_fitObjectSampling(normal_compiled_model):
    N, q, X = 100, [0.25, 0.5, 0.75], [-0.1, 0.3, 0.8]
    fit = normal_compiled_model.sampling(N, q, X)
    samples = fit.stan_obj
    assert all(fit.mu == samples.extract('mu')['mu'])
    assert all(fit.sigma == samples.extract('sigma')['sigma'])

@pytest.mark.slow
def test_fitObjectOptimizing(normal_compiled_model):
    N, q, X = 100, [0.25, 0.5, 0.75], [-0.1, 0.3, 0.8]
    fit = normal_compiled_model.optimizing(N, q, X)
    opt = fit.opt
    assert fit.mu == opt['mu']
    assert fit.sigma == opt['sigma']

@pytest.mark.slow
def test_fitObjectSampling_expected_fail(normal_compiled_model):
    N, q, X = 100, [0.25, 0.5, 0.75], [-0.1, 0.3, 0.8]
    fit = normal_compiled_model.sampling(N, q, X)
    # fits parameters are mu and sigma only
    with pytest.raises(AttributeError):
        fit.bla  # bla is not an attribute of fit

@pytest.mark.slow
def test_NormalQM_sampling(normal_compiled_model):
    N, q, X = 100, [0.25, 0.5, 0.75], [0.1, 0.3, 0.8]
    fit = normal_compiled_model.sampling(N, q, X)
    assert fit.pdf(3.) < 0.1
    assert fit.cdf(3.) > 0.9
    assert all(fit.ppf(.9) > 0.)
    assert fit.pdf(3., method='median') < 0.1
    assert fit.cdf(3., method='median') > 0.9
    assert np.all(fit.ppf(.9, method='median') > 0.)

@pytest.mark.slow
def test_NormalQM_optimizing(normal_compiled_model):
    N, q, X = 100, [0.25, 0.5, 0.75], [0.1, 0.3, 0.8]
    fit = normal_compiled_model.optimizing(N, q, X)
    assert fit.pdf(3.) < 0.1
    assert fit.cdf(3.) > 0.9
    assert fit.ppf(.9) > 0.




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bqme",
    version="0.2.0",
    author="Rajbir Singh Nirwan",
    author_email="rajbir.nirwan@gmail.com",
    description="Bayesian Quanile Matching Estimation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/RSNirwan/BQME",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        'pystan>=2.19',
        'numpy>=1.19.2',
        'scipy>=1.5.2',
    ],
    extras_require={
        "dev":[
            "pytest>=6.0",
            "pytest-cov==2.10.1",
            "sphinx==3.2.1",
            "sphinx_rtd_theme==0.5.0",
            "recommonmark==0.6.0",
        ],
    },
    package_data={
        'bqme':['stan_code_template.stan',],
    },
    include_package_data=True,
)



from typing import Dict, Tuple, List

import numpy as np
from scipy.stats import norm, gamma, lognorm, weibull_min

from bqme.variables import ContinuousVariable, PositiveContinuousVariable
from bqme.variables import Variable


class Distribution:
    """
    Base class for all distribution
    Key and values of `parameters_dict` needs the same order as the 
    input to the distribution in stan.
    """
    def __init__(self,
            parameters_dict: Dict[str, Variable],
            name: str):
        self.name = name
        self.parameters_dict = parameters_dict

    def __str__(self):
        params = {name:param.value for name, param in 
                    self.parameters_dict.items()}
        return self.__class__.__name__ + '(' +  \
            ', '.join([f'{k}={v}' for k,v in params.items()]) + \
            f', name="{self.name}"' + ')'

    def __repr__(self):
        return self.__str__()

    def domain(self): pass

    def _stan_code(self) -> Dict[str, str]:
        #real is hard coded

        #parameter initialization
        lower, upper = self.domain()
        parameter = f'real$ {self.name};'
        l = f'lower={lower}' if lower > float('-inf') else ''
        r = f'upper={upper}' if upper < float('inf') else ''
        if l and r:
            parameter = parameter.replace('$', f'<{l}, {r}>')
        elif l or r:
            parameter = parameter.replace('$', f'<{l if l else r}>')
        else:
            parameter = parameter.replace('$', '')

        #prior
        values = ", ".join([
            str(param.value) for param in self.parameters_dict.values()
            ])
        prior = f'{self.name} ~ {self.__class__.__name__.lower()}({values});'

        return {'parameter':parameter, 'prior':prior}


    def code(self) -> Dict[str, str]:
        return self._stan_code()


    def pdf(self, x:List[float]) -> np.ndarray:
        return self._distribution.pdf(x)


    def cdf(self, x:List[float]) -> np.ndarray:
        return self._distribution.cdf(x)


    def logpdf(self, x:List[float]) -> np.ndarray:
        return self._distribution.logpdf(x)

    
    def logcdf(self, x:List[float]) -> np.ndarray:
        return self._distribution.logcdf(x)


    def ppf(self, q:List[float]) -> np.ndarray:
        return self._distribution.ppf(q)
        



class Normal(Distribution):
    """
    Container for Normal Distribution

    Parameters
    ----------
    mu : float
        location of the normal
    sigma : float
        standard deviation of the normal
    name : str

    Examples
    ________
    >>> Normal(0, 1, name='mu')
    Normal(mu=0, sigma=1, name="mu")
    >>> Normal(0, 1, name='mu').code()
    {'parameter': 'real mu;', 'prior': 'mu ~ normal(0, 1);'}
    """
    def __init__(self,
            mu: float,
            sigma: float,
            name: str):
        self.mu = ContinuousVariable(mu, name='mu')
        self.sigma = PositiveContinuousVariable(sigma, name='sigma')
        self.name = name
        self._distribution = norm(loc=self.mu.value, scale=self.sigma.value)
        parameters_dict = {'mu': self.mu, 'sigma': self.sigma}
        super().__init__(parameters_dict, self.name)

    def domain(self) -> Tuple[float, float]:
        return (float('-inf'), float('inf'))



class Gamma(Distribution):
    """
    Container for Gamma Distribution

    Parameters
    ----------
    alpha : float
        shape of the Gamma
    beta : float
        rate of the Gamma
    name : str

    Examples
    ________
    >>> Gamma(1., 1., name='sigma')
    Gamma(alpha=1.0, beta=1.0, name="sigma")
    >>> Gamma(1.1, 1.0, name='sigma').code()
    {'parameter': 'real<lower=0> sigma;', 'prior': 'sigma ~ gamma(1.1, 1.0);'}
    """
    def __init__(self, 
            alpha: float,
            beta: float,
            name: str):
        self.alpha = PositiveContinuousVariable(alpha, name='alpha')
        self.beta = PositiveContinuousVariable(beta, name='beta')
        self.name = name
        self._distribution = gamma(a=self.alpha.value, scale=1./self.beta.value)
        parameters_dict = {'alpha':self.alpha, 'beta':self.beta}
        super().__init__(parameters_dict, self.name)

    def domain(self) -> Tuple[float, float]:
        return (0, float('inf'))


class Lognormal(Distribution):
    """
    Container for Lognormal Distribution

    Parameters
    ----------
    mu : float
        log(rv) has loc mu
    sigma : float
        log(rv) has scale sigma
    name : str

    Examples
    ________
    >>> Lognormal(1., 1., name='sigma')
    Lognormal(mu=1.0, sigma=1.0, name="sigma")
    >>> Lognormal(1.1, 1.0, name='sigma').code()
    {'parameter': 'real<lower=0> sigma;', 'prior': 'sigma ~ lognormal(1.1, 1.0);'}
    """
    def __init__(self,
            mu: float,
            sigma: float,
            name: str):
        self.mu = ContinuousVariable(mu, name='mu')
        self.sigma = PositiveContinuousVariable(sigma, name='sigma')
        self.name = name
        # for lognorm parameterization see scipy documentation
        self._distribution = lognorm(s=self.sigma.value, scale=np.exp(self.mu.value))
        parameters_dict = {'mu':self.mu, 'sigma':self.sigma}
        super().__init__(parameters_dict, self.name)

    def domain(self) -> Tuple[float, float]:
        return (0, float('inf'))


class Weibull(Distribution):
    """
    Container for Weibull Distribution

    Parameters
    ----------
    alpha : float
        shape of the weibull
    sigma : float
        scale of the weibull - equivalent to 1/rate
    name : str

    Examples
    ________
    >>> Weibull(1., 1., name='sigma')
    Weibull(alpha=1.0, sigma=1.0, name="sigma")
    >>> Weibull(1.1, 1.0, name='sigma').code()
    {'parameter': 'real<lower=0> sigma;', 'prior': 'sigma ~ weibull(1.1, 1.0);'}
    """
    def __init__(self, 
            alpha: float,
            sigma: float,
            name: str):
        self.alpha = PositiveContinuousVariable(alpha, name='alpha')
        self.sigma = PositiveContinuousVariable(sigma, name='sigma')
        self.name = name
        self._distribution = weibull_min(c=self.alpha.value, scale=self.sigma.value)
        parameters_dict = {'alpha':self.alpha, 'sigma':self.sigma}
        super().__init__(parameters_dict, self.name)

    def domain(self) -> Tuple[float, float]:
        return (0, float('inf'))



from typing import Dict, List

import numpy as np

class FitObject:
    """
    Base class for the fit object
    """
    def __getattr__(self, attr:str) -> 'numpy.ndarray':
        """
        allows to extract model parameters from fit object as attributes
        """
        try:
            ret = self._access_parameter(attr)
        except self._catch_error_access_parameter:
            raise AttributeError( f"Object '{self.__class__.__name__}' has no attribute '{attr}'")
        return ret

    def _get_samples(self) -> np.ndarray:
        names = self.model.parameters_dict.keys()
        # return shape (#parameters, #samples) - #samples is one for MAP
        return np.array([self._access_parameter(name) for name in names])

    def pdf(self, x:float or List[float], method='mean'):
        """
        Calculates the pdf of x using posterior samples or MAP estimate

        Parameters
        ----------
        x : float or List[float]
            points where the pdf should be evaluated
        method: str, default: 'mean'
            if sampling is used possible values are ('mean', 'median', 'full'). Return values is the mean over all samples if 'mean' is selected. Otherwise median or the full matrix is returned
            if optimizing is used return values are the pdf evaluation of the model at x.

        Returns
        -------
        ret : ndarray
        """
        f = lambda dist, param, x: dist(*param, name='a').pdf(x)
        return self._apply(f, x, method)

    def cdf(self, x:float or List[float], method='mean'):
        """
        Calculates the cdf of x using posterior samples or MAP estimate

        Parameters
        ----------
        x : float or List[float]
            points where the cdf should be evaluated
        method: str, default: 'mean'
            if sampling is used possible values are ('mean', 'median', 'full'). Return values is the mean over all samples if 'mean' is selected. Otherwise median or the full matrix is returned
            if optimizing is used return values are the pdf evaluation of the model at x.

        Returns
        -------
        ret : ndarray
        """
        f = lambda dist, param, x: dist(*param, name='a').cdf(x)
        return self._apply(f, x, method)

    def ppf(self, q:float or List[float], method='full'):
        """
        Calculates the percent point funtion (ppf) of x using posterior samples or MAP estimate

        Parameters
        ----------
        q : float or List[float]
            points where the ppf should be evaluated. Must be in range (0, 1)
        method: str, default: 'mean'
            if sampling is used possible values are ('mean', 'median', 'full'). Return values is the mean over all samples if 'mean' is selected. Otherwise median or the full matrix is returned
            if optimizing is used return values are the pdf evaluation of the model at x.

        Returns
        -------
        ret : ndarray
        """
        f = lambda dist, param, q: dist(*param, name='a').ppf(q)
        return self._apply(f, q, method)


class FitObjectSampling(FitObject):
    """
    Fit object using posterior samples of the model.
    This is an extension of the 'StanFit4Model'-type by composition.
    """
    def __init__(self, model:'QM', stan_fit_object:'StanFit4Model'):
        self.model = model
        self.stan_obj = stan_fit_object
        self._catch_error_access_parameter = ValueError

    def _access_parameter(self, attr:str) -> np.ndarray:
        return self.stan_obj.extract(attr)[attr]

    def _apply(self,
            f:'function',
            x:float or List[float],
            method: str
        ) -> np.ndarray:
        """applys pdf, cdf, ... for each sample to x"""
        posterior_samples = self._get_samples().T
        dist = self.model._distribution
        l = 1 if (type(x)==float or type(x)==int) else len(x)
        ret = np.zeros((posterior_samples.shape[0], l))
        for i, sample in enumerate(posterior_samples):
            ret[i] = f(dist, sample, x)
        if method == 'mean':
            ret = np.mean(ret, axis=0)
        elif method == 'median':
            ret = np.median(ret, axis=0)
        #else return full matrix
        return ret.squeeze()


class FitObjectOptimizing(FitObject):
    """
    Fit object using MAP estimate of the model
    """
    def __init__(self, model:'QM', opt_parameters:Dict):
        self.model = model
        self.opt = opt_parameters
        self._catch_error_access_parameter = KeyError

    def _access_parameter(self, attr:str) -> 'numpy.ndarray':
        return self.opt[attr]

    def _apply(self,
            f:'function',
            x:float or List[float],
            method: str
        ) -> np.ndarray:
        """applys pdf, cdf, ... to x"""
        map_estimate = self._get_samples()
        dist = self.model._distribution
        l = 1 if (type(x)==float or type(x)==int) else len(x)
        return f(dist, map_estimate, x)



import pytest

from bqme.distributions import Normal, Gamma, Lognormal, Weibull
from bqme.models import QM, NormalQM, GammaQM, LognormalQM, WeibullQM
from bqme._settings import BASE_DIR

FILLED_TEMPLATES_PATH = BASE_DIR / 'tests' / 'filled_templates'

def test_QM_expected_fail():
    '''input not of type Distribtion'''
    parameters_dict = {
        'a': Normal(0., 1., name='a'),
        'b': 1.   # input should be a Distribution
    }
    with pytest.raises(ValueError):
        QM(parameters_dict)


### NormalQM tests

def test_normalQM_print():
    mu = Normal(0., 1., name='mu')
    sigma = Gamma(1., 1., name='sigma')
    model = NormalQM(mu, sigma)
    assert str(model) == 'NormalQM(Normal(mu=0.0, sigma=1.0, name="mu"), Gamma(alpha=1.0, beta=1.0, name="sigma"))'
    assert repr(model) == 'NormalQM(Normal(mu=0.0, sigma=1.0, name="mu"), Gamma(alpha=1.0, beta=1.0, name="sigma"))'

def test_normalQM_template_replacements():
    mu = Normal(0., 1., name='loc')
    sigma = Gamma(1., 1., name='scale')
    model = NormalQM(mu, sigma)
    replacements = model._template_replacements()
    assert replacements['parametersnames'] == 'loc, scale'
    assert replacements['parameters'] == 'real loc;\n    real<lower=0> scale;'
    assert replacements['priors'] == \
            'loc ~ normal(0.0, 1.0);\n    scale ~ gamma(1.0, 1.0);'
    assert replacements['cdf'] == 'normal_cdf'
    assert replacements['lpdf'] == 'normal_lpdf'
    assert replacements['rng'] == 'normal_rng'

def test_normal_code():
    mu = Normal(0., 1., name='mu')
    sigma = Gamma(1., 1.2, name='sigma')
    model = NormalQM(mu, sigma)
    code = model.code
    with open(FILLED_TEMPLATES_PATH / 'os_normal.stan') as f:
        code_hard_coded = f.read()
    assert code == code_hard_coded

@pytest.mark.slow
def test_normal_sampling(normal_compiled_model):
    N, q, X = 1000, [0.25, 0.5, 0.75], [-0.1, 0.0, 0.1]
    samples = normal_compiled_model.sampling(N, q, X).stan_obj
    dic = samples.extract(['mu', 'sigma'])
    assert -0.01 < dic['mu'].mean() < 0.01

@pytest.mark.slow
def test_normal_optimizing(normal_compiled_model):
    N, q, X = 1000, [0.25, 0.5, 0.75], [-0.1, 0.0, 0.1]
    opt = normal_compiled_model.optimizing(N, q, X)
    assert -0.01 < opt.mu < 0.01


### GammaQM tests

def test_gamma_code():
    alpha = Gamma(1.0, 1.2, name='alpha')
    beta = Gamma(2.1, 2.2, name='beta')
    model = GammaQM(alpha, beta)
    code = model.code
    with open(FILLED_TEMPLATES_PATH / 'os_gamma.stan') as f:
        code_hard_coded = f.read()
    assert code == code_hard_coded

def test_gamma_check_domain_expected_fail():
    alpha = Gamma(1.0, 1.2, name='alpha')
    beta = Gamma(2.1, 2.2, name='beta')
    model = GammaQM(alpha, beta)
    N, q, X = 1000, [0.25, 0.5, 0.75], [-0.1, 1.0, 1.4] #-0.1 is invalid
    with pytest.raises(ValueError):
        model.sampling(N, q, X)
    with pytest.raises(ValueError):
        model.optimizing(N, q, X)

@pytest.mark.slow
def test_gamma_sampling(gamma_compiled_model):
    N, q, X = 1000, [0.25, 0.5, 0.75], [0.1, 1.0, 1.4]
    samples = gamma_compiled_model.sampling(N, q, X).stan_obj
    dic = samples.extract(['alpha', 'beta'])
    assert dic['alpha'].mean() > 0.

@pytest.mark.slow
def test_gamma_optimizing(gamma_compiled_model):
    N, q, X = 1000, [0.25, 0.5, 0.75], [0.1, 1.0, 1.4]
    opt = gamma_compiled_model.optimizing(N, q, X)
    assert opt.alpha > 0.

### LognormalQM tests

def test_lognormal_code():
    mu = Normal(1.0, 1.2, name='mu')
    sigma = Lognormal(2.1, 2.2, name='sigma')
    model = LognormalQM(mu, sigma)
    code = model.code
    with open(FILLED_TEMPLATES_PATH / 'os_lognormal.stan') as f:
        code_hard_coded = f.read()
    assert code == code_hard_coded

def test_lognormal_check_domain_expected_fail():
    mu = Normal(1.0, 1.2, name='mu')
    sigma = Lognormal(2.1, 2.2, name='sigma')
    model = LognormalQM(mu, sigma)
    N, q, X = 1000, [0.25, 0.5, 0.75], [-0.1, 1.0, 1.4] #-0.1 is invalid
    with pytest.raises(ValueError):
        model.sampling(N, q, X)
    with pytest.raises(ValueError):
        model.optimizing(N, q, X)

@pytest.mark.slow
def test_lognormal_sampling(lognormal_compiled_model):
    N, q, X = 1000, [0.25, 0.5, 0.75], [0.1, 1.0, 1.4]
    samples = lognormal_compiled_model.sampling(N, q, X).stan_obj
    dic = samples.extract(['mu', 'sigma'])
    assert dic['mu'].mean() > -1.

@pytest.mark.slow
def test_lognormal_optimizing(lognormal_compiled_model):
    N, q, X = 1000, [0.25, 0.5, 0.75], [0.1, 1.0, 1.4]
    opt = lognormal_compiled_model.optimizing(N, q, X)
    assert opt.mu > -1.

### WeibullQM tests

def test_weibull_code():
    alpha = Gamma(1.0, 1.2, name='alpha')
    sigma = Weibull(2.1, 2.2, name='sigma')
    model = WeibullQM(alpha, sigma)
    code = model.code
    with open(FILLED_TEMPLATES_PATH / 'os_weibull.stan') as f:
        code_hard_coded = f.read()
    assert code == code_hard_coded

def test_weibull_check_domain_expected_fail():
    alpha = Weibull(1.0, 1.2, name='alpha')
    sigma = Weibull(2.1, 2.2, name='sigma')
    model = WeibullQM(alpha, sigma)
    N, q, X = 1000, [0.25, 0.5, 0.75], [-0.1, 1.0, 1.4] #-0.1 is invalid
    with pytest.raises(ValueError):
        model.sampling(N, q, X)
    with pytest.raises(ValueError):
        model.optimizing(N, q, X)

@pytest.mark.slow
def test_weibull_sampling(weibull_compiled_model):
    N, q, X = 1000, [0.25, 0.5, 0.75], [0.1, 1.0, 1.4]
    samples = weibull_compiled_model.sampling(N, q, X).stan_obj
    dic = samples.extract(['alpha', 'sigma'])
    assert dic['alpha'].mean() > 0.

@pytest.mark.slow
def test_weibull_optimizing(weibull_compiled_model):
    N, q, X = 1000, [0.25, 0.5, 0.75], [0.1, 1.0, 1.4]
    opt = weibull_compiled_model.optimizing(N, q, X)
    assert opt.alpha > 0.
